@@ -21,11 +21,11 @@
 #include <dirent.h>
 
 #include <blkid.h>
+#include <libsmartcols.h>
 
 #include "c.h"
 #include "pathnames.h"
 #include "nls.h"
-#include "tt.h"
 #include "blkdev.h"
 #include "strutils.h"
 #include "xalloc.h"
@@ -39,6 +39,13 @@
 /* this is the default upper limit, could be modified by --nr */
 #define SLICES_MAX	256
 
+/* basic table settings */
+enum {
+	PARTX_RAW =		(1 << 0),
+	PARTX_NOHEADINGS =	(1 << 1),
+	PARTX_EXPORT =		(1 << 2),
+};
+
 /* all the columns (-o option) */
 enum {
 	COL_PARTNO,
@@ -59,6 +66,7 @@ enum {
 	ACT_LIST,
 	ACT_SHOW,
 	ACT_ADD,
+	ACT_UPD,
 	ACT_DELETE
 };
 
@@ -70,22 +78,22 @@ enum {
 struct colinfo {
 	const char	*name;	/* header */
 	double		whint;	/* width hint (N < 1 is in percent of termwidth) */
-	int		flags;	/* TT_FL_* */
+	int		flags;	/* SCOLS_FL_* */
 	const char      *help;
 };
 
 /* columns descriptions */
 struct colinfo infos[] = {
-	[COL_PARTNO]   = { "NR",    0.25, TT_FL_RIGHT, N_("partition number") },
-	[COL_START]    = { "START",   0.30, TT_FL_RIGHT, N_("start of the partition in sectors") },
-	[COL_END]      = { "END",     0.30, TT_FL_RIGHT, N_("end of the partition in sectors") },
-	[COL_SECTORS]  = { "SECTORS", 0.30, TT_FL_RIGHT, N_("number of sectors") },
-	[COL_SIZE]     = { "SIZE",    0.30, TT_FL_RIGHT, N_("human readable size") },
-	[COL_NAME]     = { "NAME",    0.30, TT_FL_TRUNC, N_("partition name") },
+	[COL_PARTNO]   = { "NR",    0.25, SCOLS_FL_RIGHT, N_("partition number") },
+	[COL_START]    = { "START",   0.30, SCOLS_FL_RIGHT, N_("start of the partition in sectors") },
+	[COL_END]      = { "END",     0.30, SCOLS_FL_RIGHT, N_("end of the partition in sectors") },
+	[COL_SECTORS]  = { "SECTORS", 0.30, SCOLS_FL_RIGHT, N_("number of sectors") },
+	[COL_SIZE]     = { "SIZE",    0.30, SCOLS_FL_RIGHT, N_("human readable size") },
+	[COL_NAME]     = { "NAME",    0.30, SCOLS_FL_TRUNC, N_("partition name") },
 	[COL_UUID]     = { "UUID",    36, 0, N_("partition UUID")},
-	[COL_SCHEME]   = { "SCHEME",  0.1, TT_FL_TRUNC, N_("partition table type (dos, gpt, ...)")},
-	[COL_FLAGS]    = { "FLAGS",   0.1, TT_FL_TRUNC, N_("partition flags")},
-	[COL_TYPE]     = { "TYPE",    1, TT_FL_RIGHT, N_("partition type hex or uuid")},
+	[COL_SCHEME]   = { "SCHEME",  0.1, SCOLS_FL_TRUNC, N_("partition table type (dos, gpt, ...)")},
+	[COL_FLAGS]    = { "FLAGS",   0.1, SCOLS_FL_TRUNC, N_("partition flags")},
+	[COL_TYPE]     = { "TYPE",    1, SCOLS_FL_RIGHT, N_("partition type (a string, a UUID, or hex)")},
 };
 
 #define NCOLS ARRAY_SIZE(infos)
@@ -120,7 +128,7 @@ static void assoc_loopdev(const char *fname)
 	rc = loopcxt_setup_device(&lc);
 
 	if (rc == -EBUSY)
-		err(EXIT_FAILURE, _("%s: failed to setup loop device"), fname);
+		err(EXIT_FAILURE, _("%s: failed to set up loop device"), fname);
 
 	loopdev = 1;
 }
@@ -233,7 +241,7 @@ static int get_max_partno(const char *disk, dev_t devno)
 		    !strcmp(d->d_name, ".."))
 			continue;
 #ifdef _DIRENT_HAVE_D_TYPE
-		if (d->d_type != DT_DIR)
+		if (d->d_type != DT_DIR && d->d_type != DT_UNKNOWN)
 			continue;
 #endif
 		if (strncmp(parent, d->d_name, strlen(parent)))
@@ -402,6 +410,99 @@ static int add_parts(int fd, const char *device,
 	return rc;
 }
 
+static void upd_parts_warnx(const char *device, int first, int last)
+{
+	if (first == last)
+		warnx(_("%s: error updating partition %d"), device, first);
+	else
+		warnx(_("%s: error updating partitions %d-%d"),
+				device, first, last);
+}
+
+static int upd_parts(int fd, const char *device, dev_t devno,
+		     blkid_partlist ls, int lower, int upper)
+{
+	int n, nparts, rc = 0, errfirst = 0, errlast = 0, err;
+	blkid_partition par;
+	uintmax_t start, size;
+
+	assert(fd >= 0);
+	assert(device);
+	assert(ls);
+
+	nparts = blkid_partlist_numof_partitions(ls);
+	if (!lower)
+		lower = 1;
+	if (!upper || lower < 0 || upper < 0) {
+		n = get_max_partno(device, devno);
+		if (!upper)
+			upper = n > nparts ? n : nparts;
+		else if (upper < 0)
+			upper = n + upper + 1;
+		if (lower < 0)
+			lower = n + lower + 1;
+	}
+	if (lower > upper) {
+		warnx(_("specified range <%d:%d> "
+			"does not make sense"), lower, upper);
+		return -1;
+	}
+
+	for (n = lower; n <= upper; n++) {
+		par = blkid_partlist_get_partition_by_partno(ls, n);
+		if (!par) {
+			if (verbose)
+				warn(_("%s: no partition #%d"), device, n);
+			continue;
+		}
+
+		start = blkid_partition_get_start(par);
+		size =  blkid_partition_get_size(par);
+		if (blkid_partition_is_extended(par))
+			/*
+			 * Let's follow the Linux kernel and reduce
+			 * DOS extended partition to 1 or 2 sectors.
+			 */
+			size = min(size, (uintmax_t) 2);
+
+		err = partx_del_partition(fd, n);
+		if (err == -1 && errno == ENXIO)
+			err = 0; /* good, it already doesn't exist */
+		if (err == -1 && errno == EBUSY)
+		{
+			/* try to resize */
+			err = partx_resize_partition(fd, n, start, size);
+			if (verbose)
+				printf(_("%s: partition #%d resized\n"), device, n);
+			if (err == 0)
+				continue;
+		}
+		if (err == 0 && partx_add_partition(fd, n, start, size) == 0) {
+			if (verbose)
+				printf(_("%s: partition #%d added\n"), device, n);
+			continue;
+		}
+
+		if (err == 0)
+			continue;
+		rc = -1;
+		if (verbose)
+			warn(_("%s: updating partition #%d failed"), device, n);
+		if (!errfirst)
+			errlast = errfirst = n;
+		else if (errlast + 1 == n)
+			errlast++;
+		else {
+			upd_parts_warnx(device, errfirst, errlast);
+			errlast = errfirst = n;
+		}
+	}
+
+	if (errfirst)
+		upd_parts_warnx(device, errfirst, errlast);
+	return rc;
+}
+
 static int list_parts(blkid_partlist ls, int lower, int upper)
 {
 	int i, nparts;
@@ -423,101 +524,100 @@ static int list_parts(blkid_partlist ls, int lower, int upper)
 		start = blkid_partition_get_start(par);
 		size =  blkid_partition_get_size(par);
 
-		printf(_("#%2d: %9ju-%9ju (%9ju sectors, %6ju MB)\n"),
+		printf(P_("#%2d: %9ju-%9ju (%9ju sector, %6ju MB)\n",
+			  "#%2d: %9ju-%9ju (%9ju sectors, %6ju MB)\n",
+			  size),
 		       n, start, start + size -1,
 		       size, (size << 9) / 1000000);
 	}
 	return 0;
 }
 
-static void add_tt_line(struct tt *tt, blkid_partition par)
+static int add_scols_line(struct libscols_table *table, blkid_partition par)
 {
-	struct tt_line *line;
-	int i;
+	struct libscols_line *line;
+	int i, rc = 0;
 
-	assert(tt);
+	assert(table);
 	assert(par);
 
-	line = tt_add_line(tt, NULL);
+	line = scols_table_new_line(table, NULL);
 	if (!line) {
 		warn(_("failed to add line to output"));
-		return;
+		return -ENOMEM;
 	}
 
 	for (i = 0; i < ncolumns; i++) {
-		char *str = NULL;
-		int rc = 0;
+		char *str = NULL;			/* allocated string */
+		const char *cstr = NULL;		/* foreign string */
 
 		switch (get_column_id(i)) {
 		case COL_PARTNO:
-			rc = xasprintf(&str, "%d",
-					blkid_partition_get_partno(par));
+			xasprintf(&str, "%d", blkid_partition_get_partno(par));
 			break;
 		case COL_START:
-			rc = xasprintf(&str, "%ju",
-					blkid_partition_get_start(par));
+			xasprintf(&str, "%ju", blkid_partition_get_start(par));
 			break;
 		case COL_END:
-			rc = xasprintf(&str, "%ju",
+			xasprintf(&str, "%ju",
 					blkid_partition_get_start(par) +
 					blkid_partition_get_size(par) - 1);
 			break;
 		case COL_SECTORS:
-			rc = xasprintf(&str, "%ju",
-					blkid_partition_get_size(par));
+			xasprintf(&str, "%ju", blkid_partition_get_size(par));
 			break;
 		case COL_SIZE:
 			if (partx_flags & FL_BYTES)
-				rc = xasprintf(&str, "%ju", (uintmax_t)
+				xasprintf(&str, "%ju", (uintmax_t)
 					blkid_partition_get_size(par) << 9);
 			else
 				str = size_to_human_string(SIZE_SUFFIX_1LETTER,
 					blkid_partition_get_size(par) << 9);
 			break;
 		case COL_NAME:
-			str = (char *) blkid_partition_get_name(par);
-			if (str)
-				str = xstrdup(str);
+			cstr = blkid_partition_get_name(par);
 			break;
 		case COL_UUID:
-			str = (char *) blkid_partition_get_uuid(par);
-			if (str)
-				str = xstrdup(str);
+			cstr = blkid_partition_get_uuid(par);
 			break;
 		case COL_TYPE:
-			str = (char *) blkid_partition_get_type_string(par);
-			if (str)
-				str = xstrdup(str);
+			if (blkid_partition_get_type_string(par))
+				cstr = blkid_partition_get_type_string(par);
 			else
-				rc = xasprintf(&str, "0x%x",
+				xasprintf(&str, "0x%x",
 					blkid_partition_get_type(par));
 			break;
 		case COL_FLAGS:
-			rc = xasprintf(&str, "0x%llx", blkid_partition_get_flags(par));
+			xasprintf(&str, "0x%llx", blkid_partition_get_flags(par));
 			break;
 		case COL_SCHEME:
 		{
 			blkid_parttable tab = blkid_partition_get_table(par);
-			if (tab) {
-				str = (char *) blkid_parttable_get_type(tab);
-				if (str)
-					str = xstrdup(str);
-			}
+			if (tab)
+				cstr = blkid_parttable_get_type(tab);
 			break;
 		}
 		default:
 			break;
 		}
 
-		if (rc || str)
-			tt_line_set_data(line, i, str);
+		if (cstr)
+			rc = scols_line_set_data(line, i, cstr);
+		else if (str)
+			rc = scols_line_refer_data(line, i, str);
+		if (rc) {
+			warn(_("failed to add data to output table"));
+			break;
+		}
 	}
+
+	return rc;
 }
 
-static int show_parts(blkid_partlist ls, int tt_flags, int lower, int upper)
+static int show_parts(blkid_partlist ls, int scols_flags, int lower, int upper)
 {
 	int i, rc = -1;
-	struct tt *tt;
+	struct libscols_table *table;
 	int nparts;
 
 	assert(ls);
@@ -526,16 +626,20 @@ static int show_parts(blkid_partlist ls, int tt_flags, int lower, int upper)
 	if (!nparts)
 		return 0;
 
-	tt = tt_new_table(tt_flags);
-	if (!tt) {
+	scols_init_debug(0);
+	table = scols_new_table();
+	if (!table) {
 		warn(_("failed to initialize output table"));
 		return -1;
 	}
+	scols_table_enable_raw(table, !!(scols_flags & PARTX_RAW));
+	scols_table_enable_export(table, !!(scols_flags & PARTX_EXPORT));
+	scols_table_enable_noheadings(table, !!(scols_flags & PARTX_NOHEADINGS));
 
 	for (i = 0; i < ncolumns; i++) {
 		struct colinfo *col = get_column_info(i);
 
-		if (!tt_define_column(tt, col->name, col->whint, col->flags)) {
+		if (!scols_table_new_column(table, col->name, col->whint, col->flags)) {
 			warnx(_("failed to initialize output column"));
 			goto done;
 		}
@@ -550,13 +654,15 @@ static int show_parts(blkid_partlist ls, int tt_flags, int lower, int upper)
 		if (upper && n > upper)
 			continue;
 
-		add_tt_line(tt, par);
+		rc = add_scols_line(table, par);
+		if (rc)
+			break;
 	}
 
 	rc = 0;
-	tt_print_table(tt);
+	scols_print_table(table);
 done:
-	tt_free_table(tt);
+	scols_unref_table(table);
 	return rc;
 }
 
@@ -604,23 +710,22 @@ static void __attribute__((__noreturn__)) usage(FILE *out)
 
 	fputs(USAGE_HEADER, out);
 	fprintf(out,
-	      _(" %s [-a|-d|-s] [--nr <n:m> | <partition>] <disk>\n"),
+	      _(" %s [-a|-d|-s|-u] [--nr <n:m> | <partition>] <disk>\n"),
 		program_invocation_short_name);
 
 	fputs(USAGE_OPTIONS, out);
-	fputs(_(" -a, --add            add specified partitions or all of them\n"
-		" -d, --delete         delete specified partitions or all of them\n"
-		" -l, --list           list partitions (DEPRECATED)\n"
-		" -s, --show           list partitions\n\n"
-
-		" -b, --bytes          print SIZE in bytes rather than in human readable format\n"
-		" -g, --noheadings     don't print headings for --show\n"
-		" -n, --nr <n:m>       specify the range of partitions (e.g. --nr 2:4)\n"
-		" -o, --output <type>  define which output columns to use\n"
-		" -P, --pairs          use key=\"value\" output format\n"
-		" -r, --raw            use raw output format\n"
-		" -t, --type <type>    specify the partition type (dos, bsd, solaris, etc.)\n"
-		" -v, --verbose        verbose mode\n"), out);
+	fputs(_(" -a, --add            add specified partitions or all of them\n"), out);
+	fputs(_(" -d, --delete         delete specified partitions or all of them\n"), out);
+	fputs(_(" -u, --update         update specified partitions or all of them\n"), out);
+	fputs(_(" -s, --show           list partitions\n\n"), out);
+	fputs(_(" -b, --bytes          print SIZE in bytes rather than in human readable format\n"), out);
+	fputs(_(" -g, --noheadings     don't print headings for --show\n"), out);
+	fputs(_(" -n, --nr <n:m>       specify the range of partitions (e.g. --nr 2:4)\n"), out);
+	fputs(_(" -o, --output <list>  define which output columns to use\n"), out);
+	fputs(_(" -P, --pairs          use key=\"value\" output format\n"), out);
+	fputs(_(" -r, --raw            use raw output format\n"), out);
+	fputs(_(" -t, --type <type>    specify the partition type (dos, bsd, solaris, etc.)\n"), out);
+	fputs(_(" -v, --verbose        verbose mode\n"), out);
 
 	fputs(USAGE_SEPARATOR, out);
 	fputs(USAGE_HELP, out);
@@ -639,7 +744,7 @@ static void __attribute__((__noreturn__)) usage(FILE *out)
 int main(int argc, char **argv)
 {
 	int fd, c, what = ACT_NONE, lower = 0, upper = 0, rc = 0;
-	int tt_flags = 0;
+	int scols_flags = 0;
 	char *type = NULL;
 	char *device = NULL; /* pointer to argv[], ie: /dev/sda1 */
 	char *wholedisk = NULL; /* allocated, ie: /dev/sda */
@@ -654,6 +759,7 @@ int main(int argc, char **argv)
 		{ "show",	no_argument,       NULL, 's' },
 		{ "add",	no_argument,       NULL, 'a' },
 		{ "delete",	no_argument,	   NULL, 'd' },
+		{ "update",     no_argument,       NULL, 'u' },
 		{ "type",	required_argument, NULL, 't' },
 		{ "nr",		required_argument, NULL, 'n' },
 		{ "output",	required_argument, NULL, 'o' },
@@ -676,7 +782,7 @@ int main(int argc, char **argv)
 	atexit(close_stdout);
 
 	while ((c = getopt_long(argc, argv,
-				"abdglrsvn:t:o:PhV", long_opts, NULL)) != -1) {
+				"abdglrsuvn:t:o:PhV", long_opts, NULL)) != -1) {
 
 		err_exclusive_options(c, long_opts, excl, excl_st);
 
@@ -691,7 +797,7 @@ int main(int argc, char **argv)
 			what = ACT_DELETE;
 			break;
 		case 'g':
-			tt_flags |= TT_FL_NOHEADINGS;
+			scols_flags |= PARTX_NOHEADINGS;
 			break;
 		case 'l':
 			what = ACT_LIST;
@@ -704,11 +810,11 @@ int main(int argc, char **argv)
 			outarg = optarg;
 			break;
 		case 'P':
-			tt_flags |= TT_FL_EXPORT;
+			scols_flags |= PARTX_EXPORT;
 			what = ACT_SHOW;
 			break;
 		case 'r':
-			tt_flags |= TT_FL_RAW;
+			scols_flags |= PARTX_RAW;
 			what = ACT_SHOW;
 			break;
 		case 's':
@@ -717,6 +823,9 @@ int main(int argc, char **argv)
 		case 't':
 			type = optarg;
 			break;
+		case 'u':
+			what = ACT_UPD;
+			break;
 		case 'v':
 			verbose = 1;
 			break;
@@ -866,7 +975,7 @@ int main(int argc, char **argv)
 
 			switch (what) {
 			case ACT_SHOW:
-				rc = show_parts(ls, tt_flags, lower, upper);
+				rc = show_parts(ls, scols_flags, lower, upper);
 				break;
 			case ACT_LIST:
 				rc = list_parts(ls, lower, upper);
@@ -874,6 +983,8 @@ int main(int argc, char **argv)
 			case ACT_ADD:
 				rc = add_parts(fd, wholedisk, ls, lower, upper);
 				break;
+			case ACT_UPD:
+				rc = upd_parts(fd, wholedisk, disk_devno, ls, lower, upper);
 			case ACT_NONE:
 				break;
 			default:
@@ -886,6 +997,8 @@ int main(int argc, char **argv)
 	if (loopdev)
 		loopcxt_deinit(&lc);
 
-	close(fd);
+	if (close_fd(fd) != 0)
+		err(EXIT_FAILURE, _("write failed"));
+
 	return rc ? EXIT_FAILURE : EXIT_SUCCESS;
 }
